import logging
import copy
from datetime import datetime
from dateutil.relativedelta import relativedelta

from spaceone.core import cache, utils
from spaceone.core.manager import BaseManager
from spaceone.cost_analysis.error import *
from spaceone.cost_analysis.model.cost_model import Cost, MonthlyCost, CostQueryHistory
from spaceone.cost_analysis.manager.data_source_rule_manager import DataSourceRuleManager

_LOGGER = logging.getLogger(__name__)


class CostManager(BaseManager):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.cost_model: Cost = self.locator.get_model('Cost')
        self.monthly_cost_model: MonthlyCost = self.locator.get_model('MonthlyCost')
        self.cost_query_history_model: CostQueryHistory = self.locator.get_model('CostQueryHistory')
        self.data_source_rule_mgr: DataSourceRuleManager = self.locator.get_manager('DataSourceRuleManager')
        self.exchange_rate_map = None

    def create_cost(self, params, execute_rollback=True):
        def _rollback(cost_vo):
            _LOGGER.info(f'[create_cost._rollback] '
                         f'Delete cost : {cost_vo.name} '
                         f'({cost_vo.cost_id})')
            cost_vo.delete()

        if 'region_code' in params and 'provider' in params:
            params['region_key'] = f'{params["provider"]}.{params["region_code"]}'

        billed_at = self._get_billed_at_from_billed_date(params['billed_date'])

        params['billed_year'] = billed_at.strftime('%Y')
        params['billed_month'] = billed_at.strftime('%Y-%m')

        params = self.data_source_rule_mgr.change_cost_data(params)

        cost_vo: Cost = self.cost_model.create(params)

        if execute_rollback:
            self.transaction.add_rollback(_rollback, cost_vo)

        return cost_vo

    def create_monthly_cost(self, params):
        return self.monthly_cost_model.create(params)

    def delete_cost(self, cost_id, domain_id):
        cost_vo: Cost = self.get_cost(cost_id, domain_id)
        self.delete_cost_by_vo(cost_vo)

    @staticmethod
    def delete_cost_by_vo(cost_vo: Cost):
        cost_vo.delete()

    def delete_cost_with_datasource(self, domain_id, data_source_id):
        _LOGGER.debug(f'[delete_cost_with_datasource] data_source_id: {data_source_id}')
        cost_vos = self.cost_model.filter(domain_id=domain_id, data_source_id=data_source_id)
        cost_vos.delete()

        monthly_cost_vos = self.monthly_cost_model.filter(domain_id=domain_id, data_source_id=data_source_id)
        monthly_cost_vos.delete()

        history_vos = self.cost_query_history_model.filter(domain_id=domain_id, data_source_id=data_source_id)
        history_vos.delete()

    def get_cost(self, cost_id, domain_id, only=None):
        return self.cost_model.get(cost_id=cost_id, domain_id=domain_id, only=only)

    def filter_costs(self, **conditions):
        return self.cost_model.filter(**conditions)

    def list_costs(self, query={}):
        return self.cost_model.query(**query)

    def stat_costs(self, query):
        return self.cost_model.stat(**query)

    def analyze_costs(self, query, target='SECONDARY_PREFERRED'):
        start = query['start']
        if len(start) == 4:
            query['start_field'] = 'billed_year'
            query['end_field'] = 'billed_year'
        elif len(start) == 7:
            query['start_field'] = 'billed_month'
            query['end_field'] = 'billed_month'

        query['target'] = target
        query['date_field'] = 'billed_date'
        return self.cost_model.analyze(**query)

    def filter_monthly_costs(self, **conditions):
        return self.monthly_cost_model.filter(**conditions)

    def list_monthly_costs(self, query={}):
        return self.monthly_cost_model.query(**query)

    def stat_monthly_costs(self, query):
        return self.monthly_cost_model.stat(**query)

    def analyze_monthly_costs(self, query, target='SECONDARY_PREFERRED'):
        start = query['start']
        if len(start) == 4:
            query['start_field'] = 'billed_year'
            query['end_field'] = 'billed_year'

        query['target'] = target
        query['date_field'] = 'billed_month'
        return self.monthly_cost_model.analyze(**query)

    def analyze_yearly_costs(self, query, target='SECONDARY_PREFERRED'):
        query['target'] = target
        query['date_field'] = 'billed_year'
        return self.monthly_cost_model.analyze(**query)

    @cache.cacheable(key='stat-costs:monthly:{domain_id}:{domain_id}:{query_hash}', expire=3600 * 24)
    def stat_monthly_costs_with_cache(self, query, query_hash, domain_id, data_source_id):
        return self.stat_monthly_costs(query)

    @cache.cacheable(key='analyze-costs:daily:{domain_id}:{data_source_id}:{query_hash}', expire=3600 * 24)
    def analyze_costs_with_cache(self, query, query_hash, domain_id, data_source_id, target='SECONDARY_PREFERRED'):
        return self.analyze_costs(query, target)

    @cache.cacheable(key='analyze-costs:monthly:{domain_id}:{data_source_id}:{query_hash}', expire=3600 * 24)
    def analyze_monthly_costs_with_cache(self, query, query_hash, domain_id, data_source_id, target='SECONDARY_PREFERRED'):
        return self.analyze_monthly_costs(query, target)

    @cache.cacheable(key='analyze-costs:yearly:{domain_id}:{data_source_id}:{query_hash}', expire=3600 * 24)
    def analyze_yearly_costs_with_cache(self, query, query_hash, domain_id, data_source_id, target='SECONDARY_PREFERRED'):
        return self.analyze_yearly_costs(query, target)

    def analyze_costs_by_granularity(self, query, domain_id, data_source_id):
        self._check_date_range(query)
        granularity = query['granularity']
        # Save query history to speed up data loading
        query_hash = utils.dict_to_hash(query)
        self.create_cost_query_history(query, query_hash, domain_id, data_source_id)

        if granularity == 'DAILY':
            return self.analyze_costs_with_cache(query, query_hash, domain_id, data_source_id)
        elif granularity == 'MONTHLY':
            return self.analyze_monthly_costs_with_cache(query, query_hash, domain_id, data_source_id)
        elif granularity == 'YEARLY':
            return self.analyze_yearly_costs_with_cache(query, query_hash, domain_id, data_source_id)

    @cache.cacheable(key='cost-query-history:{domain_id}:{data_source_id}:{query_hash}', expire=600)
    def create_cost_query_history(self, query, query_hash, domain_id, data_source_id):
        def _rollback(history_vo):
            _LOGGER.info(f'[create_cost_query_history._rollback] Delete cost query history: {query_hash}')
            history_vo.delete()

        history_model: CostQueryHistory = self.locator.get_model('CostQueryHistory')

        history_vos = history_model.filter(query_hash=query_hash, domain_id=domain_id)
        if history_vos.count() == 0:
            history_vo = history_model.create({
                'query_hash': query_hash,
                'query_options': copy.deepcopy(query),
                'data_source_id': data_source_id,
                'domain_id': domain_id
            })

            self.transaction.add_rollback(_rollback, history_vo)
        else:
            history_vos[0].update({})

    def list_cost_query_history(self, query={}):
        history_model: CostQueryHistory = self.locator.get_model('CostQueryHistory')
        return history_model.query(**query)

    @staticmethod
    def remove_stat_cache(domain_id, data_source_id):
        cache.delete_pattern(f'analyze-costs:*:{domain_id}:{data_source_id}:*')
        cache.delete_pattern(f'stat-costs:*:{domain_id}:{data_source_id}:*')
        cache.delete_pattern(f'cost-query-history:{domain_id}:{data_source_id}:*')

    def _check_date_range(self, query):
        start_str = query.get('start')
        end_str = query.get('end')
        granularity = query.get('granularity')

        start = self._parse_start_time(start_str, granularity)
        end = self._parse_end_time(end_str, granularity)
        now = datetime.utcnow().date()

        if len(start_str) != len(end_str):
            raise ERROR_INVALID_DATE_RANGE(start=start_str, end=end_str,
                                           reason='Start date and end date must be the same format.')

        if start >= end:
            raise ERROR_INVALID_DATE_RANGE(start=start_str, end=end_str,
                                           reason='End date must be greater than start date.')

        if granularity == 'DAILY':
            if start + relativedelta(months=1) < end:
                raise ERROR_INVALID_DATE_RANGE(start=start_str, end=end_str,
                                               reason='Request up to a maximum of 1 month.')

            if start + relativedelta(months=12) < now:
                raise ERROR_INVALID_DATE_RANGE(start=start_str, end=end_str,
                                               reason='For DAILY, you cannot request data older than 1 year.')

        elif granularity == 'MONTHLY':
            if start + relativedelta(months=12) < end:
                raise ERROR_INVALID_DATE_RANGE(start=start_str, end=end_str,
                                               reason='Request up to a maximum of 12 months.')

            if start + relativedelta(months=36) < now:
                raise ERROR_INVALID_DATE_RANGE(start=start_str, end=end_str,
                                               reason='For MONTHLY, you cannot request data older than 3 years.')
        elif granularity == 'YEARLY':
            if start + relativedelta(years=3) < now:
                raise ERROR_INVALID_DATE_RANGE(start=start_str, end=end_str,
                                               reason='For YEARLY, you cannot request data older than 3 years.')

    def _parse_start_time(self, date_str, granularity):
        return self._convert_date_from_string(date_str.strip(), 'start', granularity)

    def _parse_end_time(self, date_str, granularity):
        end = self._convert_date_from_string(date_str.strip(), 'end', granularity)

        if granularity == 'YEARLY':
            return end + relativedelta(years=1)
        elif granularity == 'MONTHLY':
            return end + relativedelta(months=1)
        else:
            return end + relativedelta(days=1)

    @staticmethod
    def _convert_date_from_string(date_str, key, granularity):
        if granularity == 'YEARLY':
            date_format = '%Y'
            date_type = 'YYYY'
        elif granularity == 'MONTHLY':
            if len(date_str) == 4:
                date_format = '%Y'
                date_type = 'YYYY'
            else:
                date_format = '%Y-%m'
                date_type = 'YYYY-MM'
        else:
            if len(date_str) == 4:
                date_format = '%Y'
                date_type = 'YYYY'
            elif len(date_str) == 7:
                date_format = '%Y-%m'
                date_type = 'YYYY-MM'
            else:
                date_format = '%Y-%m-%d'
                date_type = 'YYYY-MM-DD'

        try:
            return datetime.strptime(date_str, date_format).date()
        except Exception as e:
            raise ERROR_INVALID_PARAMETER_TYPE(key=key, type=date_type)

    @staticmethod
    def _get_billed_at_from_billed_date(billed_date):
        date_format = '%Y-%m-%d'

        try:
            return datetime.strptime(billed_date, date_format)
        except Exception as e:
            raise ERROR_INVALID_PARAMETER_TYPE(key='billed_date', type='YYYY-MM-DD')



import functools
from typing import List
from spaceone.api.cost_analysis.v1 import budget_pb2
from spaceone.core.pygrpc.message_type import *
from spaceone.core import utils
from spaceone.cost_analysis.model.budget_model import Budget, PlannedLimit, Notification, ProviderFilter

__all__ = ['BudgetInfo', 'BudgetsInfo']


def PlannedLimitsInfo(planned_limit_vos: List[PlannedLimit]):
    if planned_limit_vos is None:
        planned_limit_vos = []

    planned_limits_info = []

    for vo in planned_limit_vos:
        info = {
            'date': vo.date,
            'limit': vo.limit
        }

        planned_limits_info.append(budget_pb2.PlannedLimit(**info))

    return planned_limits_info


def BudgetNotificationsInfo(notification_vos: List[Notification]):
    if notification_vos is None:
        notification_vos = []

    notifications_info = []

    for vo in notification_vos:
        info = {
            'threshold': vo.threshold,
            'unit': vo.unit,
            'notification_type': vo.notification_type
        }

        notifications_info.append(budget_pb2.BudgetNotification(**info))

    return notifications_info


def ProviderFilterInfo(provider_filter_vo: ProviderFilter):
    if provider_filter_vo is None:
        return None

    info = {
        'state': provider_filter_vo.state,
        'providers': list(provider_filter_vo.providers)
    }

    return budget_pb2.ProviderFilter(**info)


def BudgetInfo(budget_vo: Budget, minimal=False):
    info = {
        'budget_id': budget_vo.budget_id,
        'name': budget_vo.name,
        'limit': budget_vo.limit,
        'currency': budget_vo.currency,
        'provider_filter': ProviderFilterInfo(budget_vo.provider_filter),
        'project_id': budget_vo.project_id,
        'project_group_id': budget_vo.project_group_id,
        'data_source_id': budget_vo.data_source_id,
    }

    if not minimal:
        info.update({
            'planned_limits': PlannedLimitsInfo(budget_vo.planned_limits),
            'time_unit': budget_vo.time_unit,
            'start': budget_vo.start,
            'end': budget_vo.end,
            'notifications': BudgetNotificationsInfo(budget_vo.notifications),
            'tags': change_struct_type(budget_vo.tags),
            'domain_id': budget_vo.domain_id,
            'created_at': utils.datetime_to_iso8601(budget_vo.created_at),
            'updated_at': utils.datetime_to_iso8601(budget_vo.updated_at)
        })

    return budget_pb2.BudgetInfo(**info)


def BudgetsInfo(budget_vos, total_count, **kwargs):
    return budget_pb2.BudgetsInfo(results=list(
        map(functools.partial(BudgetInfo, **kwargs), budget_vos)), total_count=total_count)



from mongoengine import *

from spaceone.core.model.mongo_model import MongoModel


class PluginInfo(EmbeddedDocument):
    plugin_id = StringField(max_length=40)
    version = StringField(max_length=255)
    options = DictField(default={})
    metadata = DictField(default={})
    secret_id = StringField(max_length=40, default=None, null=True)
    schema = StringField(max_length=255, default=None, null=True)
    upgrade_mode = StringField(max_length=255, choices=('AUTO', 'MANUAL'), default='AUTO')

    def to_dict(self):
        return dict(self.to_mongo())


class SecretFilter(EmbeddedDocument):
    state = StringField(max_length=20, default='ENABLED', choices=('ENABLED', 'DISABLED'))
    secrets = ListField(StringField(max_length=40), defualt=None, null=True)
    service_accounts = ListField(StringField(max_length=40), default=None, null=True)
    schemas = ListField(StringField(max_length=40), default=None, null=True)

    def to_dict(self):
        return dict(self.to_mongo())


class DataSource(MongoModel):
    data_source_id = StringField(max_length=40, generate_id='ds', unique=True)
    name = StringField(max_length=255, unique_with='domain_id')
    state = StringField(max_length=20, default='ENABLED', choices=('ENABLED', 'DISABLED'))
    data_source_type = StringField(max_length=20, choices=('LOCAL', 'EXTERNAL'))
    secret_type = StringField(max_length=32, default='MANUAL', choices=('MANUAL', 'USE_SERVICE_ACCOUNT_SECRET'))
    secret_filter = EmbeddedDocumentField(SecretFilter, default=None, null=True)
    provider = StringField(max_length=40, default=None, null=True)
    plugin_info = EmbeddedDocumentField(PluginInfo, default=None, null=True)
    template = DictField(default={})
    tags = DictField(default={})
    cost_tag_keys = ListField(StringField())
    cost_additional_info_keys = ListField(StringField())
    domain_id = StringField(max_length=40)
    created_at = DateTimeField(auto_now_add=True)
    last_synchronized_at = DateTimeField(default=None, null=True)

    meta = {
        'updatable_fields': [
            'name',
            'state',
            'plugin_info',
            'secret_filter',
            'template',
            'tags',
            'last_synchronized_at',
            'cost_tag_keys',
            'cost_additional_info_keys'
        ],
        'minimal_fields': [
            'data_source_id',
            'name',
            'state',
            'data_source_type',
            'secret_type',
            'provider'
        ],
        'ordering': [
            'name'
        ],
        'indexes': [
            'name',
            'state',
            'data_source_type',
            'provider',
            'domain_id',
            'cost_tag_keys',
            'cost_additional_info_keys',
        ]
    }



import logging

from spaceone.core.manager import BaseManager
from spaceone.cost_analysis.model.budget_model import Budget

_LOGGER = logging.getLogger(__name__)


class BudgetManager(BaseManager):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.budget_model: Budget = self.locator.get_model('Budget')

    def create_budget(self, params):
        def _rollback(budget_vo):
            _LOGGER.info(f'[create_budget._rollback] '
                         f'Delete budget : {budget_vo.name} '
                         f'({budget_vo.budget_id})')
            budget_vo.delete()

        budget_vo: Budget = self.budget_model.create(params)
        self.transaction.add_rollback(_rollback, budget_vo)

        return budget_vo

    def update_budget(self, params):
        budget_vo: Budget = self.get_budget(params['budget_id'], params['domain_id'])
        return self.update_budget_by_vo(params, budget_vo)

    def update_budget_by_vo(self, params, budget_vo):
        def _rollback(old_data):
            _LOGGER.info(f'[update_budget_by_vo._rollback] Revert Data : '
                         f'{old_data["budget_id"]}')
            budget_vo.update(old_data)

        self.transaction.add_rollback(_rollback, budget_vo.to_dict())
        return budget_vo.update(params)

    def delete_budget(self, budget_id, domain_id):
        budget_vo: Budget = self.get_budget(budget_id, domain_id)
        budget_vo.delete()

    def get_budget(self, budget_id, domain_id, only=None):
        return self.budget_model.get(budget_id=budget_id, domain_id=domain_id, only=only)

    def filter_budgets(self, **conditions):
        return self.budget_model.filter(**conditions)

    def list_budgets(self, query={}):
        return self.budget_model.query(**query)

    def stat_budgets(self, query):
        return self.budget_model.stat(**query)



import logging
from datetime import datetime
from dateutil.rrule import rrule, MONTHLY, YEARLY

from spaceone.core.service import *
from spaceone.cost_analysis.error import *
from spaceone.cost_analysis.manager.data_source_manager import DataSourceManager
from spaceone.cost_analysis.manager.budget_manager import BudgetManager
from spaceone.cost_analysis.manager.budget_usage_manager import BudgetUsageManager
from spaceone.cost_analysis.manager.identity_manager import IdentityManager
from spaceone.cost_analysis.model.budget_model import Budget
from spaceone.cost_analysis.model.data_source_model import DataSource

_LOGGER = logging.getLogger(__name__)


@authentication_handler
@authorization_handler
@mutation_handler
@event_handler
class BudgetService(BaseService):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.budget_mgr: BudgetManager = self.locator.get_manager('BudgetManager')

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['data_source_id', 'time_unit', 'start', 'end', 'domain_id'])
    def create(self, params):
        """Register budget

        Args:
            params (dict): {
                'data_source_id': 'str',
                'name': 'str',
                'project_id': 'str',
                'project_group_id': 'str',
                'limit': 'float',
                'planned_limits': 'list',
                'time_unit': 'str',
                'start': 'str',
                'end': 'str',
                'provider_filter': 'dict',
                'notifications': 'list',
                'tags': 'dict',
                'domain_id': 'str'
            }

        Returns:
            budget_vo (object)
        """

        domain_id = params['domain_id']
        data_source_id = params['data_source_id']
        project_id = params.get('project_id')
        project_group_id = params.get('project_group_id')
        limit = params.get('limit')
        planned_limits = params.get('planned_limits', [])
        time_unit = params['time_unit']
        start = params['start']
        end = params['end']
        provider_filter = params.get('provider_filter', {})
        provider_filter_state = provider_filter.get('state', 'DISABLED')
        notifications = params.get('notifications', [])

        self._check_target(project_id, project_group_id, domain_id)
        self._check_time_period(start, end)

        # Check Provider Filter
        if provider_filter_state == 'ENABLED':
            if len(provider_filter.get('providers', [])) == 0:
                raise ERROR_PROVIDER_FILTER_IS_EMPTY()
        else:
            params['provider_filter']['providers'] = []

        data_source_mgr: DataSourceManager = self.locator.get_manager('DataSourceManager')
        data_source_vo: DataSource = data_source_mgr.get_data_source(data_source_id, domain_id)
        data_source_metadata = data_source_vo.plugin_info.get('metadata', {})
        params['currency'] = data_source_metadata['currency']

        if time_unit == 'TOTAL':
            if limit is None:
                raise ERROR_REQUIRED_PARAMETER(key='limit')

            params['planned_limits'] = None

        else:
            # Check Planned Limits
            self._check_planned_limits(start, end, time_unit, planned_limits)

            params['limit'] = 0
            for planned_limit in planned_limits:
                params['limit'] += planned_limit.get('limit', 0)

        # Check Notifications
        self._check_notifications(notifications)

        budget_vo = self.budget_mgr.create_budget(params)

        # Create budget usages
        budget_usage_mgr: BudgetUsageManager = self.locator.get_manager('BudgetUsageManager')
        budget_usage_mgr.create_budget_usages(budget_vo)
        budget_usage_mgr.update_cost_usage(budget_vo.budget_id, budget_vo.domain_id)

        return budget_vo

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['budget_id', 'domain_id'])
    # @change_date_value(['end'])
    def update(self, params):
        """Update budget

        Args:
            params (dict): {
                'budget_id': 'str',
                'name': 'str',
                'limit': 'float',
                'planned_limits': 'list',
                'tags': 'dict'
                'domain_id': 'str'
            }

        Returns:
            budget_vo (object)
        """

        budget_id = params['budget_id']
        domain_id = params['domain_id']
        planned_limits = params.get('planned_limits')

        budget_usage_mgr: BudgetUsageManager = self.locator.get_manager('BudgetUsageManager')

        budget_vo: Budget = self.budget_mgr.get_budget(budget_id, domain_id)

        # Check limit and Planned Limits

        budget_vo = self.budget_mgr.update_budget_by_vo(params, budget_vo)

        if 'name' in params:
            budget_usage_vos = budget_usage_mgr.filter_budget_usages(budget_id=budget_id)
            for budget_usage_vo in budget_usage_vos:
                budget_usage_mgr.update_budget_usage_by_vo({'name': params['name']}, budget_usage_vo)

        return budget_vo

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['budget_id', 'domain_id'])
    def set_notification(self, params):
        """Set budget notification

        Args:
            params (dict): {
                'budget_id': 'str',
                'notifications': 'list',
                'domain_id': 'str'
            }

        Returns:
            budget_vo (object)
        """
        budget_id = params['budget_id']
        domain_id = params['domain_id']
        notifications = params.get('notifications', [])

        # Check Notifications
        self._check_notifications(notifications)
        params['notifications'] = notifications

        budget_vo: Budget = self.budget_mgr.get_budget(budget_id, domain_id)

        return self.budget_mgr.update_budget_by_vo(params, budget_vo)

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['budget_id', 'domain_id'])
    def delete(self, params):
        """Deregister budget

        Args:
            params (dict): {
                'budget_id': 'str',
                'domain_id': 'str'
            }

        Returns:
            None
        """

        self.budget_mgr.delete_budget(params['budget_id'], params['domain_id'])

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['budget_id', 'domain_id'])
    def get(self, params):
        """ Get budget

        Args:
            params (dict): {
                'budget_id': 'str',
                'domain_id': 'str',
                'only': 'list
            }

        Returns:
            budget_vo (object)
        """

        budget_id = params['budget_id']
        domain_id = params['domain_id']

        return self.budget_mgr.get_budget(budget_id, domain_id, params.get('only'))

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['domain_id'])
    @append_query_filter(['budget_id', 'name', 'project_id', 'project_group_id', 'data_source_id', 'domain_id'])
    @append_keyword_filter(['budget_id', 'name'])
    def list(self, params):
        """ List budgets

        Args:
            params (dict): {
                'budget_id': 'str',
                'name': 'str',
                'project_id': 'str',
                'project_group_id': 'str',
                'data_source_id': 'str',
                'domain_id': 'str',
                'query': 'dict (spaceone.api.core.v1.Query)',
                'user_projects': 'list', // from meta,
                'user_project_groups': 'list', // from meta
            }

        Returns:
            budget_vos (object)
            total_count
        """

        query = self._set_user_project_or_project_group_filter(params)
        return self.budget_mgr.list_budgets(query)

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['query', 'domain_id'])
    @append_query_filter(['domain_id'])
    @append_keyword_filter(['budget_id', 'name'])
    def stat(self, params):
        """
        Args:
            params (dict): {
                'domain_id': 'str',
                'query': 'dict (spaceone.api.core.v1.StatisticsQuery)',
                'user_projects': 'list', // from meta,
                'user_project_groups': 'list', // from meta
            }

        Returns:
            values (list) : 'list of statistics data'

        """

        query = self._set_user_project_or_project_group_filter(params)
        return self.budget_mgr.stat_budgets(query)

    def _check_target(self, project_id, project_group_id, domain_id):
        if project_id is None and project_group_id is None:
            raise ERROR_REQUIRED_PARAMETER(key='project_id or project_group_id')

        if project_id and project_group_id:
            raise ERROR_ONLY_ONF_OF_PROJECT_OR_PROJECT_GROUP()

        identity_mgr: IdentityManager = self.locator.get_manager('IdentityManager')

        if project_id:
            identity_mgr.get_project(project_id, domain_id)
        else:
            identity_mgr.get_project_group(project_group_id, domain_id)

    @staticmethod
    def _check_time_period(start, end):
        if start >= end:
            raise ERROR_INVALID_TIME_RANGE(start=start, end=end)

    def _check_planned_limits(self, start, end, time_unit, planned_limits):
        planned_limits_dict = self._convert_planned_limits_data_type(planned_limits)
        date_format = '%Y-%m'

        try:
            start_dt = datetime.strptime(start, date_format)
        except Exception as e:
            raise ERROR_INVALID_PARAMETER_TYPE(key='start', type=date_format)

        try:
            end_dt = datetime.strptime(end, date_format)
        except Exception as e:
            raise ERROR_INVALID_PARAMETER_TYPE(key='end', type=date_format)

        for dt in rrule(MONTHLY, dtstart=start_dt, until=end_dt):
            date_str = dt.strftime(date_format)
            if date_str not in planned_limits_dict:
                raise ERROR_NO_DATE_IN_PLANNED_LIMITS(date=date_str)

            del planned_limits_dict[date_str]

        if len(planned_limits_dict.keys()) > 0:
            raise ERROR_DATE_IS_WRONG(date=list(planned_limits_dict.keys()))

    @staticmethod
    def _convert_planned_limits_data_type(planned_limits):
        planned_limits_dict = {}

        for planned_limit in planned_limits:
            date = planned_limit.get('date')
            limit = planned_limit.get('limit', 0)
            if date is None:
                raise ERROR_DATE_IS_REQUIRED(value=planned_limit)

            if limit < 0:
                raise ERROR_LIMIT_IS_WRONG(value=planned_limit)

            planned_limits_dict[date] = limit

        return planned_limits_dict

    @staticmethod
    def _check_notifications(notifications):
        for notification in notifications:
            unit = notification.get('unit')
            notification_type = notification.get('notification_type')
            threshold = notification.get('threshold', 0)

            if unit not in ['PERCENT', 'ACTUAL_COST']:
                raise ERROR_UNIT_IS_REQUIRED(value=notification)

            if notification_type not in ['CRITICAL', 'WARNING']:
                raise ERROR_NOTIFICATION_TYPE_IS_REQUIRED(value=notification)

            if threshold < 0:
                raise ERROR_THRESHOLD_IS_WRONG(value=notification)

            if unit == 'PERCENT':
                if threshold > 100:
                    raise ERROR_THRESHOLD_IS_WRONG_IN_PERCENT_TYPE(value=notification)

    @staticmethod
    def _set_user_project_or_project_group_filter(params):
        query = params.get('query', {})
        query['filter'] = query.get('filter', [])

        if 'user_projects' in params:
            user_projects = params['user_projects'] + [None]
            query['filter'].append({'k': 'user_projects', 'v': user_projects, 'o': 'in'})

        if 'user_project_groups' in params:
            user_project_groups = params['user_project_groups'] + [None]
            query['filter'].append({'k': 'user_project_groups', 'v': user_project_groups, 'o': 'in'})

        return query



from mongoengine import *

from spaceone.core.model.mongo_model import MongoModel


class CostQuerySet(MongoModel):
    cost_query_set_id = StringField(max_length=40, generate_id='query', unique=True)
    name = StringField(max_length=255)
    options = DictField(default={})
    tags = DictField(default={})
    user_id = StringField(max_length=40)
    data_source_id = StringField(max_length=40)
    domain_id = StringField(max_length=40)
    created_at = DateTimeField(auto_now_add=True)
    updated_at = DateTimeField(auto_now=True)

    meta = {
        'updatable_fields': [
            'name',
            'options',
            'tags'
        ],
        'minimal_fields': [
            'cost_query_set_id',
            'name',
            'user_id',
            'data_source_id'
        ],
        'ordering': [
            'name'
        ],
        'indexes': [
            'name',
            'user_id',
            'data_source_id',
            'domain_id'
        ]
    }



import logging

from spaceone.core.service import *
from spaceone.cost_analysis.manager.cost_query_set_manager import CostQuerySetManager
from spaceone.cost_analysis.model.cost_query_set_model import CostQuerySet

_LOGGER = logging.getLogger(__name__)


@authentication_handler
@authorization_handler
@mutation_handler
@event_handler
class CostQuerySetService(BaseService):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.cost_query_set_mgr: CostQuerySetManager = self.locator.get_manager('CostQuerySetManager')

    @transaction(append_meta={'authorization.scope': 'USER'})
    @check_required(['data_source_id', 'name', 'options', 'domain_id'])
    @change_date_value(['start', 'end'])
    def create(self, params):
        """Register cost_query_set

        Args:
            params (dict): {
                'data_source_id': 'str',
                'name': 'str',
                'options': 'str',
                'tags': 'dict',
                'domain_id': 'str'
            }

        Returns:
            cost_query_set_vo (object)
        """

        params['user_id'] = self.transaction.get_meta('user_id')

        return self.cost_query_set_mgr.create_cost_query_set(params)

    @transaction(append_meta={'authorization.scope': 'USER'})
    @check_required(['cost_query_set_id', 'domain_id'])
    @change_date_value(['end'])
    def update(self, params):
        """Update cost_query_set

        Args:
            params (dict): {
                'cost_query_set_id': 'str',
                'name': 'str',
                'options': 'dict',
                'tags': 'dict'
                'domain_id': 'str'
            }

        Returns:
            cost_query_set_vo (object)
        """
        cost_query_set_id = params['cost_query_set_id']
        domain_id = params['domain_id']

        cost_query_set_vo: CostQuerySet = self.cost_query_set_mgr.get_cost_query_set(cost_query_set_id, domain_id)

        return self.cost_query_set_mgr.update_cost_query_set_by_vo(params, cost_query_set_vo)

    @transaction(append_meta={'authorization.scope': 'USER'})
    @check_required(['cost_query_set_id', 'domain_id'])
    def delete(self, params):
        """Deregister cost_query_set

        Args:
            params (dict): {
                'cost_query_set_id': 'str',
                'domain_id': 'str'
            }

        Returns:
            None
        """

        self.cost_query_set_mgr.delete_cost_query_set(params['cost_query_set_id'], params['domain_id'])

    @transaction(append_meta={'authorization.scope': 'USER'})
    @check_required(['cost_query_set_id', 'domain_id'])
    def get(self, params):
        """ Get cost_query_set

        Args:
            params (dict): {
                'cost_query_set_id': 'str',
                'domain_id': 'str',
                'only': 'list
            }

        Returns:
            cost_query_set_vo (object)
        """

        cost_query_set_id = params['cost_query_set_id']
        domain_id = params['domain_id']

        return self.cost_query_set_mgr.get_cost_query_set(cost_query_set_id, domain_id, params.get('only'))

    @transaction(append_meta={'authorization.scope': 'USER'})
    @check_required(['data_source_id', 'domain_id'])
    @append_query_filter(['data_source_id', 'cost_query_set_id', 'name', 'user_id', 'domain_id'])
    @append_keyword_filter(['cost_query_set_id', 'name'])
    def list(self, params):
        """ List cost_query_sets

        Args:
            params (dict): {
                'cost_query_set_id': 'str',
                'name': 'str',
                'user_id': 'str',
                'domain_id': 'str',
                'query': 'dict (spaceone.api.core.v1.Query)'
            }

        Returns:
            cost_query_set_vos (object)
            total_count
        """

        query = params.get('query', {})
        return self.cost_query_set_mgr.list_cost_query_sets(query)

    @transaction(append_meta={'authorization.scope': 'USER'})
    @check_required(['query', 'data_source_id', 'domain_id'])
    @append_query_filter(['data_source_id', 'domain_id'])
    @append_keyword_filter(['cost_query_set_id', 'name'])
    def stat(self, params):
        """
        Args:
            params (dict): {
                'domain_id': 'str',
                'query': 'dict (spaceone.api.core.v1.StatisticsQuery)'
            }

        Returns:
            values (list) : 'list of statistics data'

        """

        query = params.get('query', {})
        return self.cost_query_set_mgr.stat_cost_query_sets(query)



import logging
from datetime import datetime

from spaceone.core import config
from spaceone.core import utils
from spaceone.core.token import get_token
from spaceone.core.locator import Locator
from spaceone.core.scheduler import HourlyScheduler

_LOGGER = logging.getLogger(__name__)


class DataSourceSyncScheduler(HourlyScheduler):

    def __init__(self, queue, interval, minute=':00'):
        super().__init__(queue, interval, minute)
        self.locator = Locator()
        self._init_config()
        self._create_metadata()

    def _init_config(self):
        self._token = get_token('TOKEN')
        self._data_source_sync_hour = config.get_global('DATA_SOURCE_SYNC_HOUR', 16)

    def _create_metadata(self):
        self._metadata = {
            'token': self._token,
            'service': 'cost_analysis',
            'resource': 'Job',
            'verb': 'create_jobs_by_data_source'
        }

    def create_task(self):
        if datetime.utcnow().hour == self._data_source_sync_hour:
            stp = {
                'name': 'data_source_sync_schedule',
                'version': 'v1',
                'executionEngine': 'BaseWorker',
                'stages': [{
                    'locator': 'SERVICE',
                    'name': 'JobService',
                    'metadata': self._metadata,
                    'method': 'create_jobs_by_data_source',
                    'params': {
                        'params': {}
                    }
                }]
            }

            print(f'{utils.datetime_to_iso8601(datetime.now())} [INFO] [create_task] create_jobs_by_data_source => START')
            return [stp]
        else:
            print(f'{utils.datetime_to_iso8601(datetime.now())} [INFO] [create_task] create_jobs_by_data_source => SKIP')
            print(f'{utils.datetime_to_iso8601(datetime.now())} [INFO] [create_task] data_source_sync_time: {self._data_source_sync_hour} hour (UTC)')
            return []



import logging

from google.protobuf.json_format import MessageToDict

from spaceone.core.connector import BaseConnector
from spaceone.core import pygrpc
from spaceone.core.utils import parse_endpoint
from spaceone.core.error import *

__all__ = ['DataSourcePluginConnector']

_LOGGER = logging.getLogger(__name__)


class DataSourcePluginConnector(BaseConnector):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.client = None
        self.secret_data = None
        self.options = None
        self.schema = None

    def initialize(self, endpoint):
        static_endpoint = self.config.get('endpoint')

        if static_endpoint:
            endpoint = static_endpoint

        e = parse_endpoint(endpoint)
        self.client = pygrpc.client(endpoint=f'{e.get("hostname")}:{e.get("port")}', version='plugin')

        self.secret_data = self.config.get('secret_data')
        self.options = self.config.get('options')
        self.schema = self.config.get('schema')

    def init(self, options, domain_id):
        response = self.client.DataSource.init({
            'options': options,
            'domain_id': domain_id
        })

        return self._change_message(response)

    def verify(self, options, secret_data, schema, domain_id):
        params = {
            'options': self.options or options,
            'secret_data': self.secret_data or secret_data,
            'schema': self.schema or schema,
            'domain_id': domain_id
        }

        self.client.DataSource.verify(params)

    def get_tasks(self, options, secret_data, schema, domain_id, start=None, last_synchronized_at=None):
        params = {
            'options': self.options or options,
            'secret_data': self.secret_data or secret_data,
            'schema': self.schema or schema,
            'start': start,
            'last_synchronized_at': last_synchronized_at,
            'domain_id': domain_id
        }

        response = self.client.Job.get_tasks(params)
        return self._change_message(response)

    def get_cost_data(self, options, secret_data, schema, task_options, domain_id):
        params = {
            'options': self.options or options,
            'secret_data': self.secret_data or secret_data,
            'schema': self.schema or schema,
            'task_options': task_options,
            'domain_id': domain_id
        }

        response_stream = self.client.Cost.get_data(params)
        return self._process_stream(response_stream)

    def _process_stream(self, response_stream):
        for message in response_stream:
            yield self._change_message(message)

    @staticmethod
    def _change_message(message):
        return MessageToDict(message, preserving_proto_field_name=True)



import functools
from spaceone.api.cost_analysis.v1 import budget_usage_pb2
from spaceone.core.pygrpc.message_type import *
from spaceone.core import utils
from spaceone.cost_analysis.model.budget_usage_model import BudgetUsage

__all__ = ['BudgetUsageInfo', 'BudgetUsagesInfo']


def BudgetUsageInfo(budget_usage_vo: BudgetUsage, minimal=False):
    info = {
        'budget_id': budget_usage_vo.budget_id,
        'name': budget_usage_vo.name,
        'date': budget_usage_vo.date,
        'cost': budget_usage_vo.cost,
        'limit': budget_usage_vo.limit,
        'currency': budget_usage_vo.currency,
        'project_id': budget_usage_vo.project_id,
        'project_group_id': budget_usage_vo.project_group_id,
        'data_source_id': budget_usage_vo.data_source_id,
    }

    if not minimal:
        info.update({
            'domain_id': budget_usage_vo.domain_id,
            'updated_at': utils.datetime_to_iso8601(budget_usage_vo.updated_at)
        })

    return budget_usage_pb2.BudgetUsageInfo(**info)


def BudgetUsagesInfo(budget_usage_vos, total_count, **kwargs):
    return budget_usage_pb2.BudgetUsagesInfo(results=list(
        map(functools.partial(BudgetUsageInfo, **kwargs), budget_usage_vos)), total_count=total_count)



DATABASE_AUTO_CREATE_INDEX = True
DATABASES = {
    'default': {
        'db': 'cost-analysis',
        'host': 'localhost',
        'port': 27017,
        'username': '',
        'password': ''
    }
}

CACHES = {
    'default': {},
    'local': {
        'backend': 'spaceone.core.cache.local_cache.LocalCache',
        'max_size': 128,
        'ttl': 300
    }
}

HANDLERS = {
}

CONNECTORS = {
    'SpaceConnector': {
        'backend': 'spaceone.core.connector.space_connector.SpaceConnector',
        'endpoints': {
            'identity': 'grpc://identity:50051',
            'plugin': 'grpc://plugin:50051',
            'repository': 'grpc://repository:50051',
            'secret': 'grpc://secret:50051',
            'notification': 'grpc://notification:50051'

        }
    },
    'DataSourcePluginConnector': {
    },
}

# Scheduler Settings
QUEUES = {}
SCHEDULERS = {}
WORKERS = {}
TOKEN = ""
TOKEN_INFO = {}

# Job Settings
JOB_TIMEOUT = 600
DATA_SOURCE_SYNC_HOUR = 16  # Hour (UTC)
COST_QUERY_CACHE_TIME = 3   # Day

DEFAULT_EXCHANGE_RATE = {
    # 'KRW': 1178.7,
    # 'JPY': 114,
    # 'CNY': 6.3
}

INSTALLED_DATA_SOURCE_PLUGINS = [
    # {
    #     'name': '',
    #     'data_source_type': 'EXTERNAL',
    #     'plugin_info': {
    #         'plugin_id': '',
    #         'version': '',
    #         'options': {},
    #         'secret_data': {},
    #         'schema': '',
    #         'upgrade_mode': ''
    #     },
    #     'tags':{
    #         'description': ''
    #     }
    # }
]



import logging
from spaceone.core.service import *
from spaceone.cost_analysis.error import *
from spaceone.cost_analysis.service.job_service import JobService
from spaceone.cost_analysis.manager.repository_manager import RepositoryManager
from spaceone.cost_analysis.manager.secret_manager import SecretManager
from spaceone.cost_analysis.manager.data_source_plugin_manager import DataSourcePluginManager
from spaceone.cost_analysis.manager.data_source_manager import DataSourceManager
from spaceone.cost_analysis.manager.job_manager import JobManager
from spaceone.cost_analysis.manager.cost_manager import CostManager
from spaceone.cost_analysis.manager.budget_usage_manager import BudgetUsageManager
from spaceone.cost_analysis.manager.identity_manager import IdentityManager
from spaceone.cost_analysis.model.data_source_model import DataSource

_LOGGER = logging.getLogger(__name__)


@authentication_handler
@authorization_handler
@mutation_handler
@event_handler
class DataSourceService(BaseService):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.data_source_mgr: DataSourceManager = self.locator.get_manager('DataSourceManager')
        self.ds_plugin_mgr: DataSourcePluginManager = self.locator.get_manager('DataSourcePluginManager')
        self.cost_mgr: CostManager = self.locator.get_manager('CostManager')
        self.budget_usage_mgr: BudgetUsageManager = self.locator.get_manager('BudgetUsageManager')
        self.job_mgr: JobManager = self.locator.get_manager('JobManager')

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['name', 'data_source_type', 'domain_id'])
    def register(self, params):
        """Register data source

        Args:
            params (dict): {
                'name': 'str',
                'data_source_type': 'str',
                'secret_type': 'str',
                'secret_filter': 'dict',
                'provider': 'str',
                'template': 'dict',
                'plugin_info': 'dict',
                'tags': 'dict',
                'domain_id': 'str'
            }

        Returns:
            data_source_vo (object)
        """

        domain_id = params['domain_id']
        data_source_type = params['data_source_type']

        if data_source_type == 'EXTERNAL':
            params['template'] = None

            plugin_info = params.get('plugin_info', {})
            secret_type = params.get('secret_type', 'MANUAL')

            if secret_type == 'USE_SERVICE_ACCOUNT_SECRET' and 'provider' not in params:
                raise ERROR_REQUIRED_PARAMETER(key='provider')

            self._validate_plugin_info(plugin_info, secret_type)
            self._check_plugin(plugin_info['plugin_id'], domain_id)

            if 'secret_filter' in params:
                self.validate_secret_filter(params['secret_filter'], params['domain_id'])

            # Update metadata
            endpoint, updated_version = self.ds_plugin_mgr.get_data_source_plugin_endpoint(plugin_info, domain_id)
            if updated_version:
                params['plugin_info']['version'] = updated_version

            options = params['plugin_info'].get('options', {})

            plugin_metadata = self._init_plugin(endpoint, options, domain_id)

            params['plugin_info']['metadata'] = plugin_metadata

            secret_data = plugin_info.get('secret_data')
            if secret_type == 'MANUAL' and secret_data:
                self._verify_plugin(endpoint, plugin_info, domain_id)

                secret_mgr: SecretManager = self.locator.get_manager('SecretManager')
                secret_id = secret_mgr.create_secret(domain_id, secret_data, plugin_info.get('schema'))

                params['plugin_info']['secret_id'] = secret_id
                del params['plugin_info']['secret_data']

        else:
            params['plugin_info'] = None
            params['secret_type'] = None
            params['secret_filter'] = None

            if template := params.get('template'):
                # Check Template
                pass
            else:
                raise ERROR_REQUIRED_PARAMETER(key='template')

        data_source_vo: DataSource = self.data_source_mgr.register_data_source(params)

        if data_source_type == 'EXTERNAL':
            data_source_id = data_source_vo.data_source_id
            metadata = data_source_vo.plugin_info.metadata

            self.ds_plugin_mgr.create_data_source_rules_by_metadata(metadata, data_source_id, domain_id)

        return data_source_vo

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['data_source_id', 'domain_id'])
    def update(self, params):
        """Update data source

        Args:
            params (dict): {
                'data_source_id': 'str',
                'name': 'str',
                'secret_filter': 'dict',
                'template': 'dict',
                'tags': 'dict'
                'domain_id': 'str'
            }

        Returns:
            data_source_vo (object)
        """
        data_source_id = params['data_source_id']
        domain_id = params['domain_id']
        data_source_vo: DataSource = self.data_source_mgr.get_data_source(data_source_id, domain_id)

        if 'secret_filter' in params:
            if data_source_vo.secret_type == 'USE_SERVICE_ACCOUNT_SECRET':
                self.validate_secret_filter(params['secret_filter'], params['domain_id'])
            else:
                raise ERROR_NOT_ALLOW_SECRET_FILTER(data_source_id=data_source_id)

        if 'template' in params:
            if data_source_vo.data_source_type == 'LOCAL':
                # Check Template
                pass
            else:
                raise ERROR_NOT_ALLOW_PLUGIN_SETTINGS(data_source_id=data_source_id)

        return self.data_source_mgr.update_data_source_by_vo(params, data_source_vo)

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['data_source_id', 'domain_id'])
    def enable(self, params):
        """ Enable data source

        Args:
            params (dict): {
                'data_source_id': 'str',
                'domain_id': 'str'
            }

        Returns:
            data_source_vo (object)
        """

        data_source_id = params['data_source_id']
        domain_id = params['domain_id']
        data_source_vo: DataSource = self.data_source_mgr.get_data_source(data_source_id, domain_id)

        return self.data_source_mgr.update_data_source_by_vo({'state': 'ENABLED'}, data_source_vo)

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['data_source_id', 'domain_id'])
    def disable(self, params):
        """ Disable data source

        Args:
            params (dict): {
                'data_source_id': 'str',
                'domain_id': 'str'
            }

        Returns:
            data_source_vo (object)
        """

        data_source_id = params['data_source_id']
        domain_id = params['domain_id']
        data_source_vo: DataSource = self.data_source_mgr.get_data_source(data_source_id, domain_id)

        return self.data_source_mgr.update_data_source_by_vo({'state': 'DISABLED'}, data_source_vo)

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['data_source_id', 'domain_id'])
    def deregister(self, params):
        """Deregister data source

        Args:
            params (dict): {
                'data_source_id': 'str',
                'cascade_delete_cost: 'bool',
                'domain_id': 'str'
            }

        Returns:
            None
        """

        data_source_id = params['data_source_id']
        cascade_delete_cost = params.get('cascade_delete_cost', True)
        domain_id = params['domain_id']

        data_source_vo: DataSource = self.data_source_mgr.get_data_source(data_source_id, domain_id)

        if cascade_delete_cost:
            self.cost_mgr.delete_cost_with_datasource(domain_id, data_source_id)
            self.budget_usage_mgr.update_budget_usage(domain_id, data_source_id)
            self.cost_mgr.remove_stat_cache(domain_id, data_source_id)
            self.job_mgr.preload_cost_stat_queries(domain_id, data_source_id)

        if data_source_vo.plugin_info:
            secret_id = data_source_vo.plugin_info.secret_id

            if secret_id:
                secret_mgr: SecretManager = self.locator.get_manager('SecretManager')
                secret_mgr.delete_secret(secret_id, domain_id)

        self.data_source_mgr.deregister_data_source_by_vo(data_source_vo)

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['data_source_id', 'domain_id'])
    def sync(self, params):
        """Sync data with data source

        Args:
            params (dict): {
                'data_source_id': 'str',
                'start': 'datetime',
                'no_preload_cache': 'bool',
                'domain_id': 'str'
            }

        Returns:
            None
        """
        job_service: JobService = self.locator.get_service('JobService')

        data_source_id = params['data_source_id']
        domain_id = params['domain_id']
        job_options = {
            'no_preload_cache': params.get('no_preload_cache', False),
            'start': params.get('start')
        }

        data_source_vo: DataSource = self.data_source_mgr.get_data_source(data_source_id, domain_id)

        if data_source_vo.state == 'DISABLED':
            raise ERROR_DATA_SOURCE_STATE(data_source_id=data_source_id)

        if data_source_vo.data_source_type == 'LOCAL':
            raise ERROR_NOT_ALLOW_SYNC_COMMAND(data_source_id=data_source_id)

        return job_service.create_cost_job(data_source_vo, job_options)

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['data_source_id', 'domain_id'])
    def verify_plugin(self, params):
        """ Verify data source plugin

        Args:
            params (dict): {
                'data_source_id': 'str',
                'domain_id': 'str'
            }

        Returns:
            data_source_vo (object)
        """

        data_source_id = params['data_source_id']
        domain_id = params['domain_id']
        data_source_vo: DataSource = self.data_source_mgr.get_data_source(data_source_id, domain_id)

        if data_source_vo.data_source_type == 'LOCAL':
            raise ERROR_NOT_ALLOW_PLUGIN_SETTINGS(data_source_id=data_source_id)

        endpoint = self.ds_plugin_mgr.get_data_source_plugin_endpoint_by_vo(data_source_vo)
        plugin_info = data_source_vo.plugin_info.to_dict()

        self._verify_plugin(endpoint, plugin_info, domain_id)

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['data_source_id', 'domain_id'])
    def update_plugin(self, params):
        """Update data source plugin

        Args:
            params (dict): {
                'data_source_id': 'str',
                'version': 'str',
                'options': 'dict',
                'upgrade_mode': 'str',
                'domain_id': 'str'
            }

        Returns:
            data_source_vo (object)
        """

        data_source_id = params['data_source_id']
        domain_id = params['domain_id']
        version = params.get('version')
        options = params.get('options')
        upgrade_mode = params.get('upgrade_mode')

        data_source_vo = self.data_source_mgr.get_data_source(data_source_id, domain_id)

        if data_source_vo.data_source_type == 'LOCAL':
            raise ERROR_NOT_ALLOW_PLUGIN_SETTINGS(data_source_id=data_source_id)

        plugin_info = data_source_vo.plugin_info.to_dict()

        if version:
            plugin_info['version'] = version

        if isinstance(options, dict):
            plugin_info['options'] = options

        if upgrade_mode:
            plugin_info['upgrade_mode'] = upgrade_mode

        endpoint, updated_version = self.ds_plugin_mgr.get_data_source_plugin_endpoint(plugin_info, domain_id)
        if updated_version:
            plugin_info['version'] = updated_version

        options = plugin_info.get('options', {})
        plugin_metadata = self._init_plugin(endpoint, options, domain_id)
        plugin_info['metadata'] = plugin_metadata

        params = {
            'plugin_info': plugin_info
        }

        data_source_vo = self.data_source_mgr.update_data_source_by_vo(params, data_source_vo)

        self.ds_plugin_mgr.delete_data_source_rules(data_source_id, domain_id)
        self.ds_plugin_mgr.create_data_source_rules_by_metadata(plugin_metadata, data_source_id, domain_id)

        return data_source_vo

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['data_source_id', 'domain_id'])
    def get(self, params):
        """ Get data source

        Args:
            params (dict): {
                'data_source_id': 'str',
                'domain_id': 'str',
                'only': 'list
            }

        Returns:
            data_source_vo (object)
        """

        data_source_id = params['data_source_id']
        domain_id = params['domain_id']

        return self.data_source_mgr.get_data_source(data_source_id, domain_id, params.get('only'))

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['domain_id'])
    @append_query_filter(['data_source_id', 'name', 'state', 'data_source_type', 'provider', 'domain_id'])
    @change_tag_filter('tags')
    @append_keyword_filter(['data_source_id', 'name'])
    def list(self, params):
        """ List data sources

        Args:
            params (dict): {
                'data_source_id': 'str',
                'name': 'str',
                'state': 'str',
                'data_source_type': 'str',
                'provider': 'str',
                'domain_id': 'str',
                'query': 'dict (spaceone.api.core.v1.Query)'
            }

        Returns:
            data_source_vos (object)
            total_count
        """

        query = params.get('query', {})
        return self.data_source_mgr.list_data_sources(query)

    @transaction(append_meta={'authorization.scope': 'DOMAIN'})
    @check_required(['query', 'domain_id'])
    @append_query_filter(['domain_id'])
    @change_tag_filter('tags')
    @append_keyword_filter(['data_source_id', 'name'])
    def stat(self, params):
        """
        Args:
            params (dict): {
                'domain_id': 'str',
                'query': 'dict (spaceone.api.core.v1.StatisticsQuery)'
            }

        Returns:
            values (list) : 'list of statistics data'

        """

        query = params.get('query', {})
        return self.data_source_mgr.stat_data_sources(query)

    def validate_secret_filter(self, secret_filter, domain_id):
        if 'secrets' in secret_filter:
            _query = {'filter': [{'k': 'secret_id', 'v': secret_filter['secrets'], 'o': 'in'}]}
            secret_mgr: SecretManager = self.locator.get_manager(SecretManager)
            response = secret_mgr.list_secrets(_query, domain_id)
            if response.get('total_count', 0) != len(secret_filter['secrets']):
                raise ERROR_INVALID_PARAMETER(key='secret_filter.secrets', reason='Secrets not found')

        if 'service_accounts' in secret_filter:
            _query = {'filter': [{'k': 'service_account_id', 'v': secret_filter['service_accounts'], 'o': 'in'}]}
            identity_mgr: IdentityManager = self.locator.get_manager(IdentityManager)
            response = identity_mgr.list_service_accounts(_query, domain_id)
            if response.get('total_count', 0) != len(secret_filter['service_accounts']):
                raise ERROR_INVALID_PARAMETER(key='secret_filter.service_accounts', reason='Service accounts not found')

        if 'schemas' in secret_filter:
            _query = {'filter': [{'k': 'name', 'v': secret_filter['schemas'], 'o': 'in'}]}
            repo_mgr: RepositoryManager = self.locator.get_manager(RepositoryManager)
            response = repo_mgr.list_schemas(_query, domain_id)
            if response.get('total_count', 0) != len(secret_filter['schemas']):
                raise ERROR_INVALID_PARAMETER(key='secret_filter.schema', reason='Schema not found')

    def _check_plugin(self, plugin_id, domain_id):
        repo_mgr: RepositoryManager = self.locator.get_manager('RepositoryManager')
        repo_mgr.get_plugin(plugin_id, domain_id)

    def _init_plugin(self, endpoint, options, domain_id):
        self.ds_plugin_mgr.initialize(endpoint)
        return self.ds_plugin_mgr.init_plugin(options, domain_id)

    def _verify_plugin(self, endpoint, plugin_info, domain_id):
        options = plugin_info.get('options', {})
        secret_id = plugin_info.get('secret_id')
        secret_data = plugin_info.get('secret_data')
        schema = plugin_info.get('schema')

        if not secret_data:
            secret_data = self._get_secret_data(secret_id, domain_id)

        self.ds_plugin_mgr.initialize(endpoint)
        self.ds_plugin_mgr.verify_plugin(options, secret_data, schema, domain_id)

    def _get_secret_data(self, secret_id, domain_id):
        secret_mgr: SecretManager = self.locator.get_manager('SecretManager')
        if secret_id:
            secret_data = secret_mgr.get_secret_data(secret_id, domain_id)
        else:
            secret_data = {}

        return secret_data

    @staticmethod
    def _validate_plugin_info(plugin_info, secret_type):
        if 'plugin_id' not in plugin_info:
            raise ERROR_REQUIRED_PARAMETER(key='plugin_info.plugin_id')

        if plugin_info.get('upgrade_mode', 'AUTO') == 'MANUAL' and 'version' not in plugin_info:
            raise ERROR_REQUIRED_PARAMETER(key='plugin_info.version')

        if secret_type == 'MANUAL' and plugin_info.get('secret_data') is None:
            raise ERROR_REQUIRED_PARAMETER(key='plugin_info.secret_data')



import functools
from spaceone.api.cost_analysis.v1 import job_task_pb2
from spaceone.core.pygrpc.message_type import *
from spaceone.core import utils
from spaceone.cost_analysis.model.job_task_model import JobTask

__all__ = ['JobTaskInfo', 'JobTasksInfo']


def JobTaskInfo(job_task_vo: JobTask, minimal=False):
    info = {
        'job_task_id': job_task_vo.job_task_id,
        'status': job_task_vo.status,
        'created_count': job_task_vo.created_count,
        'job_id': job_task_vo.job_id,
        'data_source_id': job_task_vo.data_source_id
    }

    if not minimal:
        info.update({
            'options': change_struct_type(job_task_vo.options),
            'error_code': job_task_vo.error_code,
            'error_message': job_task_vo.error_message,
            'job_id': job_task_vo.job_id,
            'domain_id': job_task_vo.domain_id,
            'created_at': utils.datetime_to_iso8601(job_task_vo.created_at),
            'started_at': utils.datetime_to_iso8601(job_task_vo.started_at),
            'updated_at': utils.datetime_to_iso8601(job_task_vo.updated_at),
            'finished_at': utils.datetime_to_iso8601(job_task_vo.finished_at)
        })

    return job_task_pb2.JobTaskInfo(**info)


def JobTasksInfo(job_task_vos, total_count, **kwargs):
    return job_task_pb2.JobTasksInfo(results=list(
        map(functools.partial(JobTaskInfo, **kwargs), job_task_vos)), total_count=total_count)
